/****************************************************************************
 *
 * lflist.h -- lock-free singly linked list for grid-based matching
 *
 * This file is part of DDM
 *
 * Copyright (C) 2017 Moreno Marzolla
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#ifndef LFLIST_H
#define LFLIST_H

namespace lockfree {
    /**
     * This is an extremely basic implementation of a lock-free singly
     * linked list. The list only implements the operations required
     * by the parallel grid-based matching algorithm. The list
     * supports concurrent insertions at the front in time O(1) *
     * congestion operations. Removal of elements is not allowed. A
     * const_iterator is provided for iterating through the list.
     */
    template <typename T> struct lfnode {
        T item;
        struct lfnode<T>* next;        
        lfnode(const T& val) : item(val), next(0) { };
    };

    template <typename T> class lfiterator {
    private:
        lfnode<T> *n;
    public:
        lfiterator() : n(0) { }
        lfiterator(lfnode<T> *node) : n(node) { };
        virtual ~lfiterator() { }
        lfiterator& operator++() { n = n->next; return *this; }
        T &operator*() const { return n->item; }
        bool operator==(const lfiterator& other) { return (n == other.n); }
        bool operator!=(const lfiterator& other) { return (n != other.n); }
    };

    template <typename T> class const_lfiterator {
    private:
        const lfnode<T> *n;
    public:
        const_lfiterator() : n(0) { }
        const_lfiterator(lfnode<T> *node) : n(node) { };
        virtual ~const_lfiterator() { }
        const_lfiterator& operator++() { n = n->next; return *this; }
        const T &operator*() const { return n->item; }
        bool operator==(const const_lfiterator& other) { return (n == other.n); }
        bool operator!=(const const_lfiterator& other) { return (n != other.n); }
    };

    template<typename T> class list {
    private:
        struct lfnode<T> *head;
    public:
        typedef lfiterator<T> iterator;
        typedef const_lfiterator<T> const_iterator;

        list( ) : head(0) { };
        virtual ~list( )
        {
            struct lfnode<T> *cur = head, *next;
            while ( cur ) {
                next = cur->next;
                delete cur;
                cur = next;
            }
        }
        
        void push_front( const T& val )
        {
            lfnode<T> *new_node = new lfnode<T>(val);
            lfnode<T> *old_head;
            do {
                old_head = head;
                new_node->item = val;
                new_node->next = old_head;
            } while ( !__sync_bool_compare_and_swap(&head, old_head, new_node) );
        }
        
        //        iterator begin() const { return lfiterator<T>(head); }
        //        iterator end() const { return lfiterator<T>(); }

        const_iterator begin() const { return const_lfiterator<T>(head); }
        const_iterator end() const { return const_lfiterator<T>(); }

    };
    
} // namespace lockfree

#endif
