/****************************************************************************
 *
 * main.c
 *
 * This file is part of DDM
 *
 * Copyright (C) 2013, 2016, 2017, 2018 Moreno Marzolla
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#ifdef _OPENMP
#include <omp.h>
#endif
#include <unistd.h>
#include "timing.h"
#include "interval.h"
#include "parallel_brute_force.h"
#include "parallel_grid_matching.hh"
#include "parallel_sort_matching.hh"
#include "interval_tree_matching.h"
#include "cflags.h"

extern const char* CFLAGS;

const double Lmax = 1.0e6; /* total length */

/**
 * allocates and returns a vector of |n| random intervals of length
 * |l|. The lower bound of each interval is randomly placed in [0,
 * Lmax-l). The caller owns the returned pointer, and is therefore
 * responsible for freeing the allocated memory.
 */
struct interval* mk_random_intervals( size_t n, double l )
{
    struct interval* result = (struct interval*)malloc( n * sizeof(result[0]) );
    size_t i;
    assert( l > 0 );
    for ( i=0; i<n; ++i ) {
	result[i].lower = (Lmax-l)*random() / ((double)RAND_MAX + 1);
	result[i].upper = result[i].lower + l;
        assert( result[i].lower < result[i].upper );
        result[i].payload = NULL;
    }
    return result;
}

typedef size_t (*ddm_matching_t)( const struct interval* subscriptions, size_t n,
                                  const struct interval* updates, size_t m );


int main( int argc, char* argv[] )
{
    char fname[256];
    int opt;
    const size_t default_extents = 100000;
    const size_t default_d = 1;
    const double default_alpha = 0.01;
    const size_t default_grid_size = 100;

    size_t tot_extents = default_extents;
    size_t d = default_d; /* number of dimensions, must be 1 */
    size_t u; /* updates */
    size_t s; /* subscriptions */
    double alpha = default_alpha; /* overlap ratio */
    unsigned int rng_seed = clock();
    int i;

    /* Enumerate all supported DDM algorithms; the default is the
       first one. The last entry must be {NULL, NULL, NULL} */
    struct {
        const char *name;
        const char* description;
        ddm_matching_t fun;
    } ddm_algos[] = { {"bfm", "Parallel brute-force matching", parallel_brute_force},
                      {"itm", "Parallel Interval-Tree matching", interval_tree_matching},
                      {"sbm", "Parallel sort-based matching", parallel_sort_matching},
                      {"gbm", "Parallel grid matching", parallel_grid_matching},
                      {NULL, NULL, NULL} };

    const char *ddm_algo_name = ddm_algos[0].name;
    ddm_matching_t ddm_algo_fun = ddm_algos[0].fun;

    grid_size = default_grid_size;

    while ((opt = getopt(argc, argv, "hs:n:d:a:g:r:")) != -1) {
        switch (opt) {
        case 's':
            ddm_algo_name = NULL;
            ddm_algo_fun = NULL;
            for (i=0; ddm_algos[i].name; i++) {
                if ( 0 == strcmp(optarg, ddm_algos[i].name) ) {
                    ddm_algo_name = ddm_algos[i].name;
                    ddm_algo_fun = ddm_algos[i].fun;
                }
            }
            if ( NULL == ddm_algo_fun ) {
                fprintf(stderr, "\nFATAL: invalid algorithm %s\n", optarg);
                exit(EXIT_FAILURE);
            }
            break;
        case 'n':
            tot_extents = atol(optarg);
            break;
        case 'd':
            d = atoi(optarg);
            if ( d != 1 ) {
                fprintf(stderr, "This program only supports 1 dimension\n");
                exit(EXIT_FAILURE);
            }
            break;
        case 'a':
            alpha = atof(optarg);
            if (alpha <= 0.0) {
                fprintf(stderr, "Invalid alpha (%f): must be strictly positive\n", alpha);
                exit(EXIT_FAILURE);
            }
            break;
        case 'g':
            grid_size = atol(optarg);
            break;
        case 'r':
            rng_seed = atoi(optarg);
            break;
        default:
            fprintf(stderr, "\nUsage: %s [-s algo] [-n extents] [-d dims] [-a alpha] [-r] [-h]\n\nwhere:\n\n-s algo\t\talgorithm (see below for the list of supported algorithms)\n-n extents\tnumber of extents (default %zu)\n-d dims\t\tnumber of dimensions (the only allowed value is 1)\n-a alpha\toverlapping degree alpha (default %f)\n-g grid_size\tgrid size for grid-based matching (default %zu)\n-r val\t\tseed of the RNG (default is the current timestamp)\n-h\t\tshow this help message\n\nList of supported algorithms:\n\n", argv[0], default_extents, default_alpha, default_grid_size);
            for (i=0; ddm_algos[i].name; i++) {
                fprintf(stderr, "\t%s\t%s%s\n", ddm_algos[i].name, ddm_algos[i].description, i == 0 ? "[default]" : "");
            }
            fprintf(stderr, "\n");
            exit(EXIT_FAILURE);
        }
    }

    printf("# %s compiled with:\n", argv[0]);
    printf("# \tCFLAGS=%s\n", CFLAGS);
    printf("# \tCXXFLAGS=%s\n", CXXFLAGS);
    printf("# \tLDFLAGS=%s\n", LDFLAGS);
    printf("# \tLDLIBS=%s\n", LDLIBS);
#ifdef _OPENMP
    printf("# OPENMP enabled with max %d threads\n", omp_get_max_threads());
#endif
    printf("# Algorithm ........... %s\n", ddm_algo_name);
    if ( ddm_algo_fun == parallel_grid_matching ) {
        printf("# Grid size ........... %zu\n", grid_size);
    }
    printf("# Number of extents.... %zu\n", tot_extents);
    printf("# Number of dimensions. %zu\n", d);
    printf("# alpha................ %f\n", alpha);
    printf("# RNG seed ............ %u\n", rng_seed);

    srandom(rng_seed);
    u = s = tot_extents / 2;
    double l = alpha * Lmax / (double)(u+s);
    assert( l > 0 );
    struct interval* upds = mk_random_intervals( u, l );
    struct interval* subs = mk_random_intervals( s, l );
    struct timer timing;

    timing_init( &timing );

    timing_start( &timing );
    size_t nmatches = ddm_algo_fun( subs, s, upds, u );
    timing_stop( &timing );

    double total_time = timing_get_average( &timing );

    /* write results */
    snprintf(fname, sizeof(fname), "%s.txt", ddm_algo_name);
    FILE* fout = fopen(fname, "a");
    if ( fout == NULL ) {
	fprintf(stderr, "Error appending to file %s\n", fname);
	exit(-1);
    }

    fprintf(fout, "%f\n", total_time);
    fclose(fout);

    printf("%zu matches %f\n", nmatches, total_time );

    return 0;
}
