/****************************************************************************
 *
 * parallel_grid_matching.cc - parallel grid-based matching
 *
 * This file is part of DDM
 *
 * Copyright (C) 2017, 2018 Moreno Marzolla
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/
#include <iostream>
#include <cstdio>
#include <vector>
#include <fstream>
#include <cmath> // for floor()

#ifdef _OPENMP
#include <parallel/algorithm>
using __gnu_parallel::min_element;
using __gnu_parallel::max_element;
#else 
#include <algorithm> 
using std::min_element;
using std::max_element;
#endif

#include "parallel_grid_matching.hh"
#include "interval.h"

/**
 * Use the -DLFLIST compiler flag to use the lock-free singly linked
 * list implementation; default is use the standard std::list
 * structure and protect updates during the construction of the grid
 * using a "#pragma omp critical" directive.
 */
#ifdef LFLIST
#include "lflist.hh"
using lockfree::list;
#else
#include <list>
using std::list;
#endif

size_t grid_size;

/* Comparison function for lower bound of intervals |a| and |b| */
static int interval_compare_lower(const struct interval &a, const struct interval &b)
{
    return (a.lower < b.lower);
}

/* Comparison function for upper bound of intervals |a| and |b| */
static int interval_compare_upper(const struct interval &a, const struct interval &b)
{
    return (a.upper < b.upper);
}

/**
 * Parallel grid-based matching.
 */
size_t parallel_grid_matching( const struct interval* sub, size_t n,
                               const struct interval* upd, size_t m )
{
    size_t result = 0;

    /* compute the span of intervals; if compiling with OpenMP, this
       step uses the parallel vesione of min_element() and
       max_element() */
    const interval* lower_sub = min_element(sub, sub+n, interval_compare_lower);
    const interval* lower_upd = min_element(upd, upd+m, interval_compare_lower);
    const double bbox_lower = std::min( lower_sub->lower, lower_upd->lower );

    const interval* upper_sub = max_element(sub, sub+n, interval_compare_upper);
    const interval* upper_upd = max_element(upd, upd+m, interval_compare_upper);
    const double bbox_upper = std::max( upper_sub->upper, upper_upd->upper );

    const double grid_step = (bbox_upper - bbox_lower) / grid_size;

    std::vector< list<int> > sub_grid(grid_size);

    // Fill the grid with subscription intervals
#pragma omp parallel for default(none) shared(n,sub,upd,sub_grid,grid_size)
    for (size_t i=0; i<n; i++) {
        size_t idx = floor((sub[i].lower - bbox_lower) / grid_step);
        double grid_lower = bbox_lower + idx*grid_step;
        while( idx < grid_size && grid_lower < sub[i].upper ) {            
#ifndef LFLIST
            /* if we are not using the lock-free list implementation,
               we must protect the insertion with a mutex */
#pragma omp critical
#endif
            {
                sub_grid[idx].push_front(i);             
            }
            idx++;
            grid_lower += grid_step;
        }
    }

#if 0
    /* dump the number of elements in each cell (use for debugging) */
    {
        std::ofstream out("grid-bins.txt");
        std::vector< list<int> >::const_iterator it;
        for (it = sub_grid.begin(); it != sub_grid.end(); ++it ) {
            out << it->size() << std::endl;
        }
    }
#endif
    std::cout << "Initialization completed" << std::endl;
    
    // Match update intervals
#pragma omp parallel for default(none) shared(n,m,sub,upd,sub_grid,grid_size) reduction(+:result)
    for (size_t j=0; j<m; j++) {
        /* found[i] is true iff we have found an intersection between
           update region j and subscription region i; we must use this
           array to avoid reporting the same intersection multiple
           time */
        std::vector<bool> found(n, false); 
        size_t idx = floor((upd[j].lower - bbox_lower) / grid_step);
        double grid_lower = bbox_lower + idx*grid_step;
        // iterate over the grid cells overlapping with update region upd[j]
        while( idx < grid_size && grid_lower < upd[j].upper ) {
            const list<int> &sub_list = sub_grid[idx];
            /* iterate over all subscription regions overlapping with
               the grid cell sub_grid[idx] */
            for( list<int>::const_iterator it = sub_list.begin(); it != sub_list.end(); ++it) {
                const int sub_id = *it;
                if ( !found[sub_id] && intersect(&sub[sub_id], &upd[j]) ) {
                    found[sub_id] = true;
                    result++;
                }
            }
            idx++;
            grid_lower += grid_step;
        }        
    }

    return result;
}
